@@ -173,11 +173,6 @@ public:
 	// Kasper: explicit CTOR does not allow e.g. 'AD<AD<double> > x = 0;'
 #define CTOR_FROM(TYPE) inline AD(const TYPE &t);
   CTOR_FROM(double)
-  CTOR_FROM(int)
-  CTOR_FROM(size_t)
-#ifdef EIGEN_DEFAULT_DENSE_INDEX_TYPE_REDEFINED
-  CTOR_FROM(EIGEN_DEFAULT_DENSE_INDEX_TYPE)
-#endif
 #undef CTOR_FROM
 
 	// assignment from some other type
